@@ -1,4 +1,4 @@
-# $Id: chess.tcl 1354 2008-01-28 17:39:01Z sergei $
+# $Id: chess.tcl 1486 2008-08-08 16:31:36Z sergei $
 
 package require msgcat
 
@@ -296,6 +296,8 @@ proc chess::make_default_position {gid} {
     set flags(position,draw) 0
     set flags(position,halfmove) 0
     set flags(position,history) {}
+
+    set flags(position,repetitions) [list [code_position $gid] 1]
 }
 
 proc chess::save_position {gid} {
@@ -316,6 +318,56 @@ proc chess::restore_position {gid} {
     find_legal_moves $gid $flags(position,turn)
 }
 
+proc chess::code_position {gid} {
+    variable $gid
+    upvar 0 $gid flags
+
+    set code ""
+    for {set c 0} {$c < 8} {incr c} {
+	for {set r 0} {$r < 8} {incr r} {
+	    switch -- $flags(position,$c,$r) {
+		wk {append code K}
+		wq {append code Q}
+		wr {append code R}
+		wb {append code B}
+		wn {append code N}
+		wp {append code P}
+		bk {append code k}
+		bq {append code q}
+		br {append code r}
+		bb {append code b}
+		bn {append code n}
+		bp {append code p}
+		default {append code .}
+	    }
+	}
+    }
+
+    append code $flags(position,wk,kcastling)
+    append code $flags(position,wk,qcastling)
+    append code $flags(position,bk,kcastling)
+    append code $flags(position,bk,qcastling)
+
+    if {[llength $flags(position,enpassant)] > 0} {
+	lassign $flags(position,enpassant) c r
+	set cp1 [expr {$c + 1}]
+	set cm1 [expr {$c - 1}]
+	if {$r == 2} {
+	    if {($cp1 < 8 && $flags(position,$cp1,3) == "bp") || \
+		    ($cm1 >= 0 && $flags(position,$cm1,3) == "bp")} {
+		append code $c$r
+	    }
+	} elseif {$r == 5} {
+	    if {($cp1 < 8 && $flags(position,$cp1,4) == "wp") || \
+		    ($cm1 >= 0 && $flags(position,$cm1,4) == "wp")} {
+		append code $c$r
+	    }
+	}
+    }
+
+    return $code
+}
+
 proc chess::make_gid {jid id} {
     jid_to_tag [concat $jid $id]
 }
@@ -926,14 +978,14 @@ proc chess::do_move {gid cf rf ct rt prom draw} {
 
     if {$rt == (7 - $r)} {
 	if {$ct == 0} {
-	    set flags(position,${opp},qcastling) 0
+	    set flags(position,${opp}k,qcastling) 0
 	} elseif {$ct == 7} {
-	    set flags(position,${opp},kcastling) 0
+	    set flags(position,${opp}k,kcastling) 0
 	}
     }
     if {$flags(position,$ct,$rt) == "${p}k"} {
- 	set flags(position,${p},kcastling) 0
- 	set flags(position,${p},qcastling) 0
+ 	set flags(position,${p}k,kcastling) 0
+ 	set flags(position,${p}k,qcastling) 0
  	if {($ct - $cf) == 2} {
  	    set opts {0-0}
  	    set flags(position,5,$r) "${p}r"
@@ -945,9 +997,9 @@ proc chess::do_move {gid cf rf ct rt prom draw} {
  	}
     } elseif {$flags(position,$ct,$rt) == "${p}r"} {
  	if {($cf == 0) && ($rf == $r)} {
- 	    set flags(position,${p},qcastling) 0
+ 	    set flags(position,${p}k,qcastling) 0
  	} elseif {($cf == 7) && ($rf == $r)} {
- 	    set flags(position,${p},kcastling) 0
+ 	    set flags(position,${p}k,kcastling) 0
  	}
     }
 
@@ -1023,6 +1075,9 @@ proc chess::do_move {gid cf rf ct rt prom draw} {
     }
     lappend opts $prm $suffix
     add_move_to_history $gid $cf $rf $ct $rt $f $opts
+
+    set repetitions [add_to_repetitions $gid]
+
     if {$draw && !$my_move} {
 	attention_message $gid \
 	    [::msgcat::mc "\n\n Opponent proposes a draw\n\n"]
@@ -1033,8 +1088,11 @@ proc chess::do_move {gid cf rf ct rt prom draw} {
     }
 
     if {$stalemate} {
-	# Draw
+	# Draw by stalemate
 	end_game $gid 0.5 [::msgcat::mc "Draw (Stalemate)"]
+    } elseif {$repetitions >= 3} {
+	# Draw by repetition
+	end_game $gid 0.5 [::msgcat::mc "Draw (Repetition)"]
     } elseif {$checkmate} {
 	if {$my_move} {
 	    # I win
@@ -1186,7 +1244,7 @@ proc chess::send_move {gid cf rf ct rt prom} {
 }
 
 proc chess::send_result {gid res child} {
-    if {$res != "OK"} {
+    if {$res == "ERR"} {
 	attention_message $gid \
 	    [format [::msgcat::mc "\n\n Opponent rejected move:\n %s\n\n"] \
 		[error_to_string $child]]
@@ -1206,6 +1264,21 @@ proc chess::make_move_tag {gid cf rf ct rt prom} {
     }
 }
 
+proc chess::add_to_repetitions {gid} {
+    variable $gid
+    upvar 0 $gid flags
+
+    array set repetitions $flags(position,repetitions)
+    set code [code_position $gid]
+    if {[info exists repetitions($code)]} {
+	incr repetitions($code)
+    } else {
+	set repetitions($code) 1
+    }
+    set flags(position,repetitions) [array get repetitions]
+    return $repetitions($code)
+}
+
 proc chess::add_move_to_history {gid cf rf ct rt f opts} {
     variable piece_name
     variable $gid
