@@ -1,4 +1,4 @@
-# $Id: renju.tcl 1400 2008-04-16 13:26:14Z sergei $
+# $Id: renju.tcl 1481 2008-08-08 10:48:29Z sergei $
 
 package require msgcat
 
@@ -988,7 +988,7 @@ proc renju::send_move {gid ct rt} {
 }
 
 proc renju::send_result {gid res child} {
-    if {$res != "OK"} {
+    if {$res == "ERR"} {
 	attention_message $gid \
 	    [format [::msgcat::mc "\n\n Opponent rejected move:\n %s\n\n"] \
 		[error_to_string $child]]
