@@ -1,4 +1,4 @@
-# $Id: reversi.tcl 1354 2008-01-28 17:39:01Z sergei $
+# $Id: reversi.tcl 1481 2008-08-08 10:48:29Z sergei $
 
 package require msgcat
 
@@ -960,7 +960,7 @@ proc reversi::send_move {gid ct rt} {
 }
 
 proc reversi::send_result {gid res child} {
-    if {$res != "OK"} {
+    if {$res == "ERR"} {
 	attention_message $gid \
 	    [format [::msgcat::mc "\n\n Opponent rejected move:\n %s\n\n"] \
 		[error_to_string $child]]
