@@ -1,4 +1,4 @@
-# $Id: whiteboard.tcl 1341 2007-12-31 14:15:38Z sergei $
+# $Id: whiteboard.tcl 1493 2008-08-26 10:32:18Z sergei $
 
 uplevel #0 [list source [file join [file dirname [info script]] svgrender.tcl]]
 
@@ -62,49 +62,56 @@ proc wb::open_wb {connid jid args} {
     set tb [frame $w.tb]
     pack $tb -side left -fill y
 
-    set tbpolyline [radiobutton $w.tb.line -text [::msgcat::mc "PolyLine"] \
-			-variable [namespace current]::tool($chatid) \
-			-value polyline \
-			-command [list [namespace current]::line_bind $c $chatid]]
-    
     set tbfreehand [radiobutton $w.tb.freehand -text [::msgcat::mc "FreeHand"] \
 			-variable [namespace current]::tool($chatid) \
 			-value freehand \
+			-anchor w \
 			-command [list [namespace current]::freehand_bind \
 				      $c $chatid]]
 
-    set tbcircle [radiobutton $w.tb.circle -text [::msgcat::mc "Circle"] \
+    set tbpolyline [radiobutton $w.tb.line -text [::msgcat::mc "PolyLine"] \
 			-variable [namespace current]::tool($chatid) \
-			-value circle \
-			-command [list [namespace current]::circle_bind \
+			-value polyline \
+			-anchor w \
+			-command [list [namespace current]::line_bind $c $chatid]]
+    
+    set tbrectangle [radiobutton $w.tb.rectangle -text [::msgcat::mc "Rectangle"] \
+			-variable [namespace current]::tool($chatid) \
+			-value rectangle \
+			-anchor w \
+			-command [list [namespace current]::rectangle_bind \
 				      $c $chatid]]
 
     set tbpolygon [radiobutton $w.tb.polygon -text [::msgcat::mc "Polygon"] \
 			-variable [namespace current]::tool($chatid) \
 			-value polygon \
+			-anchor w \
 			-command [list [namespace current]::polygon_bind \
 				      $c $chatid]]
 
+    set tbcircle [radiobutton $w.tb.circle -text [::msgcat::mc "Circle"] \
+			-variable [namespace current]::tool($chatid) \
+			-value circle \
+			-anchor w \
+			-command [list [namespace current]::circle_bind \
+				      $c $chatid]]
+
     set tbtext [radiobutton $w.tb.text -text [::msgcat::mc "Text"] \
-		      -variable [namespace current]::tool($chatid) \
-		      -value text \
-		      -command [list [namespace current]::text_bind \
-				    $c $chatid]]
+			-variable [namespace current]::tool($chatid) \
+			-value text \
+			-anchor w \
+			-command [list [namespace current]::text_bind \
+				       $c $chatid]]
 
-    set tbmove [radiobutton $w.tb.move -text [::msgcat::mc "Move"] \
-		    -variable [namespace current]::tool($chatid) \
-		    -value move \
-		    -command [list [namespace current]::move_bind \
-				  $c $chatid]]
+    set tbmove [radiobutton $w.tb.move -text [::msgcat::mc "Move/Transform"] \
+			-variable [namespace current]::tool($chatid) \
+			-value move \
+			-anchor w \
+			-command [list [namespace current]::move_bind \
+				       $c $chatid]]
 
-    set tbremove [radiobutton $w.tb.remove -text [::msgcat::mc "Remove"] \
-		      -variable [namespace current]::tool($chatid) \
-		      -value remove \
-		      -command [list [namespace current]::remove_bind \
-				    $c $chatid]]
-
-    pack $tbpolyline $tbfreehand $tbcircle $tbpolygon $tbtext $tbmove $tbremove \
-	 -anchor w
+    pack $tbfreehand $tbpolyline $tbrectangle $tbpolygon $tbcircle $tbtext \
+	 $tbmove -anchor w -fill x
 
     button $w.tb.clear -text [::msgcat::mc "Clear"] \
 	-command [list [namespace current]::send_clear $chatid]
@@ -190,8 +197,8 @@ proc wb::open_wb {connid jid args} {
 	     $w.tb.color [namespace current]::width($chatid)]
 
     variable tool
-    set tool($chatid) polyline
-    line_bind $c $chatid
+    set tool($chatid) move
+    move_bind $c $chatid
 }
 
 proc wb::balloon {chatid c action X Y} {
@@ -202,14 +209,14 @@ proc wb::balloon {chatid c action X Y} {
     set id [$c find withtag current]
     set tags {}
     foreach t [$c gettags $id] {
-	if {[llength $t] == 2} {
+	if {[lindex $t 0] == "time"} {
 	    lappend tags $t
 	}
     }
 
     set msgs {}
-    foreach t [lsort -index 0 -integer $tags] {
-	lappend msgs [lindex $t 1]
+    foreach t [lsort -index 1 -integer $tags] {
+	lappend msgs [lindex $t 2]
     }
 
     balloon::default_balloon $c:$id $action $X $Y -text [join $msgs "\n"]
@@ -331,8 +338,9 @@ proc wb::handle_wb {chatid from type body x} {
 
     foreach xelem $x {
 	jlib::wrapper:splitxml $xelem tag vars isempty chdata children
-	
-	if {[cequal [jlib::wrapper:getattr $vars xmlns] tkabber:whiteboard]} {
+
+	if {[string equal [jlib::wrapper:getattr $vars xmlns] \
+			  tkabber:whiteboard]} {
 	    open_wb [chat::get_connid $chatid] [chat::get_jid $chatid]
 	    set w [win_id whiteboard $chatid]
 	    foreach child $children {
@@ -353,23 +361,55 @@ proc wb::parse_item {chatid from seconds item} {
     switch -- $tag {
 	svg {
 	    foreach child $children {
-		set id [svg::parse_svg_item $w.c $child]
+		set id [svg::parseSVGItem $w.c {} {} $child]
 		if {$id != ""} {
+		    $w.c addtag [list tag $child] withtag $id
 		    $w.c addtag [time_tag created $from $seconds] withtag $id
 		}
 	    }
 	}
+	transform {
+	    set id [jlib::wrapper:getattr $vars id]
+	    set transform [jlib::wrapper:getattr $vars transform]
+	    set transform1 $transform
+	    set tags [$w.c gettags [list id $id]]
+	    set child {}
+	    foreach t $tags {
+		if {[lindex $t 0] == "tag"} {
+		    set child [lindex $t 1]
+		    break
+		}
+	    }
+	    foreach t $tags {
+		if {[lindex $t 0] == "transform"} {
+		    set transform "$transform [lindex $t 1]"
+		    break
+		}
+	    }
+	    if {$child != {}} {
+		$w.c delete [list id $id]
+		set id1 [svg::parseSVGItem $w.c [svg::ParseTransform $transform] {} $child]
+		if {$id1 != ""} {
+		    foreach t $tags {
+			$w.c addtag $t withtag $id1
+		    }
+		    add_transform_tag $w.c $id $transform1
+		    $w.c addtag [time_tag transformed $from $seconds] withtag $id1
+		}
+	    }
+	}
 	move {
 	    set id [jlib::wrapper:getattr $vars id]
 	    set dx [jlib::wrapper:getattr $vars dx]
 	    set dy [jlib::wrapper:getattr $vars dy]
 	    if {![string is double $dx] || $dx == ""} {set dx 0}
 	    if {![string is double $dy] || $dy == ""} {set dy 0}
-	    $w.c addtag [time_tag moved $from $seconds] withtag id$id
-	    $w.c move id$id $dx $dy
+	    add_transform_tag $w.c $id translate($dx,$dy)
+	    $w.c addtag [time_tag moved $from $seconds] withtag [list id $id]
+	    $w.c move [list id $id] $dx $dy
 	}
 	remove {
-	    $w.c delete "id[jlib::wrapper:getattr $vars id]"
+	    $w.c delete [list id [jlib::wrapper:getattr $vars id]]
 	}
 	clear {
 	    $w.c delete all
@@ -379,6 +419,18 @@ proc wb::parse_item {chatid from seconds item} {
 
 ###############################################################################
 
+proc wb::add_transform_tag {c id transform} {
+    set tags [$c gettags [list id $id]]
+    foreach t $tags {
+	if {[lindex $t 0] == "transform"} {
+	    set transform "$transform [lindex $t 1]"
+	    $c dtag [list id $id] $t
+	    break
+	}
+    }
+    $c addtag [list transform $transform] withtag [list id $id]
+}
+
 proc wb::time_tag {type jid {seconds ""}} {
     set seconds_now [clock seconds]
     set format $::plugins::options(timestamp_format)
@@ -393,10 +445,13 @@ proc wb::time_tag {type jid {seconds ""}} {
     set time [clock format $seconds -format $format]
     switch -- $type {
 	created {
-	    return [list $seconds [::msgcat::mc "%s created: %s" $time $jid]]
+	    return [list time $seconds [::msgcat::mc "%s created: %s" $time $jid]]
 	}
 	moved {
-	    return [list $seconds [::msgcat::mc "%s moved: %s" $time $jid]]
+	    return [list time $seconds [::msgcat::mc "%s moved: %s" $time $jid]]
+	}
+	transformed {
+	    return [list time $seconds [::msgcat::mc "%s transformed: %s" $time $jid]]
 	}
 	default {
 	    return ""
@@ -404,37 +459,139 @@ proc wb::time_tag {type jid {seconds ""}} {
     }
 }
 
+###############################################################################
+
+proc wb::popup_menu {c chatid x y} {
+    set m .whiteboard_popup_menu
+
+    if {[winfo exists $m]} {
+	destroy $m
+    }
+
+    set tags [$c gettags current]
+    set id ""
+    foreach t $tags {
+	if {[lindex $t 0] == "id"} {
+	    set id [lindex $t 1]
+	    break
+	}
+    }
+
+    menu $m -tearoff 0
+
+    if {![string equal $id ""]} {
+	set state normal
+    } else {
+	set state disabled
+    }
+
+    $m add command -label [::msgcat::mc "Flip horizontally"] \
+		   -command [list [namespace current]::flip_h $c $chatid $id] \
+		   -state $state
+    $m add command -label [::msgcat::mc "Flip vertically"] \
+		   -command [list [namespace current]::flip_v $c $chatid $id] \
+		   -state $state
+    $m add command -label [::msgcat::mc "Rotate 45\u00b0"] \
+		   -command [list [namespace current]::rotate $c $chatid $id -45] \
+		   -state $state
+    $m add command -label [::msgcat::mc "Rotate 90\u00b0"] \
+		   -command [list [namespace current]::rotate $c $chatid $id -90] \
+		   -state $state
+    $m add command -label [::msgcat::mc "Rotate 135\u00b0"] \
+		   -command [list [namespace current]::rotate $c $chatid $id -135] \
+		   -state $state
+    $m add command -label [::msgcat::mc "Rotate 180\u00b0"] \
+		   -command [list [namespace current]::rotate $c $chatid $id -180] \
+		   -state $state
+
+    $m add separator
+
+    $m add command -label [::msgcat::mc "Remove"] \
+		   -command [list [namespace current]::remove_b1p $c $chatid $id] \
+		   -state $state
+
+    tk_popup $m $x $y
+}
+
+###############################################################################
+
+proc wb::flip_h {c chatid id} {
+    lassign [$c bbox [list id $id]] x1 y1 x2 y2
+    set x02 [expr {$x1+$x2}]
+    transform $c $chatid $id "translate($x02,0) scale(-1,1)"
+}
+
+proc wb::flip_v {c chatid id} {
+    lassign [$c bbox [list id $id]] x1 y1 x2 y2
+    set y02 [expr {$y1+$y2}]
+    transform $c $chatid $id "translate(0,$y02) scale(1,-1)"
+}
+
+proc wb::rotate {c chatid id angle} {
+    lassign [$c bbox [list id $id]] x1 y1 x2 y2
+    set x0 [expr {($x1+$x2)/2}]
+    set y0 [expr {($y1+$y2)/2}]
+    transform $c $chatid $id "rotate($angle,$x0,$y0)"
+}
+
+proc wb::transform {c chatid id transform} {
+    if {[chat::is_groupchat $chatid]} {
+	set type groupchat
+    } else {
+	set type chat
+	set jid [jlib::connection_jid [chat::get_connid $chatid]]
+	add_transform_tag $c $id $transform
+	$c addtag [time_tag transformed $jid] withtag [list id $id]
+    }
+
+    set vars [list id $id transform $transform]
+
+    set connid [chat::get_connid $chatid]
+    set jid [chat::get_jid $chatid]
+    jlib::send_msg $jid \
+	-connection $connid \
+	-type $type \
+	-xlist [list [jlib::wrapper:createtag x \
+			  -vars {xmlns tkabber:whiteboard} \
+			  -subtags [list [jlib::wrapper:createtag transform \
+					      -vars $vars]]]]
+}
+
+###############################################################################
+
 ###############################################################################
 # Line
 
-proc wb::line_bind {c jid} {
+proc wb::line_bind {c chatid} {
     bind $c <ButtonPress-1> \
-	[list [namespace current]::line_b1 [double% $c] [double% $jid] %x %y]
+	[list [namespace current]::line_b1 [double% $c] [double% $chatid] %x %y]
     bind $c <B1-Motion> {}
     bind $c <Motion> [list [namespace current]::line_b1m [double% $c] %x %y]
     bind $c <ButtonRelease-1> {}
     bind $c <Button-3> [list [namespace current]::line_b3 \
-			    [double% $c] [double% $jid]]
+			    [double% $c] [double% $chatid]]
+    $c configure -cursor crosshair
 }
 
-proc wb::line_b1 {c jid x y} {
+proc wb::line_b1 {c chatid x y} {
     variable line
 
     set x [$c canvasx $x]
     set y [$c canvasy $y]
-    #puts "$x $y"
 
     if {[info exists line(drawed)]} {
 	lappend line(coords) $x $y
-
 	catch {$c delete $line(temp)}
-	set line(temp) [eval $c create line $line(coords) $line(options)]
+	set tag [line_tag ""]
+        set line(temp) [svg::parseSVGItem $c {} {} $tag]
     } else {
 	set line(drawed) 1
-	set line(coords) "$x $y"
-	set line(options) [list -fill [get_color $jid] \
-			       -width [get_width $jid] \
-			       -joinstyle miter]
+	set line(coords) [list $x $y]
+	set line(options) [list stroke-linejoin miter stroke [get_color $chatid]]
+	if {[set width [get_width $chatid]] != 1} {
+	    lappend line(options) stroke-width $width
+	}
+
     }
 }
 
@@ -445,10 +602,12 @@ proc wb::line_b1m {c x y} {
     set y [$c canvasy $y]
 
     if {[info exists line(drawed)]} {
-	#lappend line(coords) $x $y
-
+	set coords $line(coords)
+	lappend line(coords) $x $y
 	catch {$c delete $line(temp)}
-	set line(temp) [eval $c create line $line(coords) $x $y $line(options)]
+	set tag [line_tag ""]
+        set line(temp) [svg::parseSVGItem $c {} {} $tag]
+	set line(coords) $coords
     }
 }
 
@@ -461,49 +620,52 @@ proc wb::line_b3 {c chatid} {
 
 	set id [create_id]
 	catch {$c delete $line(temp)}
-	set line(temp) [eval $c create line $line(coords) $line(options) \
-				-tag id$id]
+	set tag [line_tag $id]
+        set line(temp) [svg::parseSVGItem $c {} {} $tag]
 
 	if {[chat::is_groupchat $chatid]} {
 	    $c delete $line(temp)
 	} else {
 	    set jid [jlib::connection_jid [chat::get_connid $chatid]]
+	    $c addtag [list tag $tag] withtag $line(temp)
 	    $c addtag [time_tag created $jid] withtag $line(temp)
 	}
 
-	lappend vars points $line(coords)
-	if {[set color [get_color $chatid]] != "#000000"} {
-	    lappend vars stroke $color
-	}
-	if {[set width [get_width $chatid]] != 1} {
-	    lappend vars stroke-width $width
-	}
-
-	lappend vars id $id
-
-	send_svg $chatid [jlib::wrapper:createtag polyline -vars $vars]
+	send_svg $chatid $tag
 
 	set line(coords) {}
 	set line(temp) {}
     }
 }
 
+proc wb::line_tag {id} {
+    variable line
+
+    set vars $line(options)
+    lappend vars points $line(coords)
+    if {$id != ""} {
+	lappend vars id $id
+    }
+    return [jlib::wrapper:createtag polyline -vars $vars]
+}
+
 ###############################################################################
 
 ###############################################################################
 # Polygon
 
-proc wb::polygon_bind {c jid} {
+proc wb::polygon_bind {c chatid} {
     bind $c <ButtonPress-1> \
-	[list [namespace current]::polygon_b1 [double% $c] [double% $jid] %x %y]
+	[list [namespace current]::polygon_b1 [double% $c] [double% $chatid] %x %y]
     bind $c <B1-Motion> {}
     bind $c <Motion> [list [namespace current]::polygon_m [double% $c] %x %y]
     bind $c <ButtonRelease-1> {}
     bind $c <Button-3> [list [namespace current]::polygon_b3 \
-			    [double% $c] [double% $jid]]
+			    [double% $c] [double% $chatid]]
+    $c configure -cursor crosshair
 }
 
-proc wb::polygon_b1 {c jid x y} {
+proc wb::polygon_b1 {c chatid x y} {
     variable polygon
     variable line1
     variable line2
@@ -518,24 +680,21 @@ proc wb::polygon_b1 {c jid x y} {
 	catch {$c delete $line1(temp)}
 	catch {$c delete $line2(temp)}
 	catch {$c delete $polygon(temp)}
-	if {$dofill == 1} {
-	    set polygon(temp) [eval $c create polygon $polygon(coords) $polygon(options)]
-	} else {
-	    set polygon(temp) [eval $c create line $polygon(coords) \
-				       [lindex $polygon(coords) 0] \
-				       [lindex $polygon(coords) 1] $polygon(line_options)]
-	}
+	set tag [polygon_tag ""]
+        set polygon(temp) [svg::parseSVGItem $c {} {} $tag]
     } else {
 	set polygon(drawed) 1
-	set polygon(coords) "$x $y"
-	set polygon(line_options) [list -fill [get_color $jid] \
-			       -width [get_width $jid] \
+	set polygon(coords) [list $x $y]
+	set polygon(line_options) [list -fill [get_color $chatid] \
+			       -width [get_width $chatid] \
 			       -joinstyle miter]
-	set polygon(options) [list -fill [get_fill_color $jid] \
-			       -outline [get_color $jid] \
-			       -width [get_width $jid] \
-			       -joinstyle miter]
-	
+	set polygon(options) [list stroke-linejoin miter stroke [get_color $chatid]]
+	if {[set width [get_width $chatid]] != 1} {
+	    lappend polygon(options) stroke-width $width
+	}
+	if {$dofill == 1} {
+	    lappend polygon(options) fill [get_fill_color $chatid]
+	}
     }
 }
 
@@ -574,123 +733,209 @@ proc wb::polygon_b3 {c chatid} {
 	catch {$c delete $line1(temp)}
 	catch {$c delete $line2(temp)}
 	catch {$c delete $polygon(temp)}
+
+	set tag [polygon_tag $id]
+
 	if {[llength $polygon(coords)] > 4} {
-	    if {$dofill == 1} {
-		set polygon(temp) \
-		    [eval $c create polygon $polygon(coords) $polygon(options) \
-			     -tag id$id]
-	    } else {
-	    	set polygon(temp) \
-		    [eval $c create line $polygon(coords) \
-			     [lindex $polygon(coords) 0] \
-			     [lindex $polygon(coords) 1] \
-			     $polygon(line_options) \
-		             -tag id$id]
-	    }
+	    set polygon(temp) [svg::parseSVGItem $c {} {} $tag]
 
 	    if {[chat::is_groupchat $chatid]} {
 		$c delete $polygon(temp)
 	    } else {
 		set jid [jlib::connection_jid [chat::get_connid $chatid]]
+		$c addtag [list tag $tag] withtag $polygon(temp)
 		$c addtag [time_tag created $jid] withtag $polygon(temp)
 	    }
 
-	    lappend vars points $polygon(coords)
-	    if {$dofill == 1} {
-		lappend vars fill [get_fill_color $chatid]
-	    }
-	    lappend vars stroke [get_color $chatid]
-	    if {[set width [get_width $chatid]] != 1} {
-		lappend vars stroke-width $width
-	    }
-
-	    lappend vars id $id
-
-	    send_svg $chatid [jlib::wrapper:createtag polygon -vars $vars]
+	    send_svg $chatid $tag
 	}
 	set polygon(coords) {}
 	set polygon(temp) {}
     }
 }
 
+proc wb::polygon_tag {id} {
+    variable polygon
+
+    set vars $polygon(options)
+    lappend vars points $polygon(coords)
+    if {$id != ""} {
+	lappend vars id $id
+    }
+    return [jlib::wrapper:createtag polygon -vars $vars]
+}
+
+###############################################################################
+
+###############################################################################
+# Rectangle
+
+proc wb::rectangle_bind {c chatid} {
+    bind $c <ButtonPress-1> \
+	[list [namespace current]::rectangle_b1 [double% $c] [double% $chatid] %x %y]
+    bind $c <B1-Motion> \
+ 	[list [namespace current]::rectangle_b1m [double% $c] %x %y]
+    bind $c <Motion> {}
+    bind $c <ButtonRelease-1> \
+	[list [namespace current]::rectangle_b1r [double% $c] [double% $chatid] %x %y]
+    bind $c <Button-3> {}
+    $c configure -cursor crosshair
+}
+
+proc wb::rectangle_b1 {c chatid x y} {
+    variable rectangle
+    global dofill
+
+    set x [$c canvasx $x]
+    set y [$c canvasy $y]
+
+    set rectangle(drawed) 1
+    set rectangle(x1) $x
+    set rectangle(y1) $y
+    set rectangle(options) [list stroke [get_color $chatid]]
+    if {$dofill == 1} {
+	lappend rectangle(options) fill [get_fill_color $chatid]
+    }
+    if {[set width [get_width $chatid]] != 1} {
+	lappend rectangle(options) stroke-width $width
+    }
+}
+
+proc wb::rectangle_b1r {c chatid x y} {
+    variable rectangle
+
+    if {[info exists rectangle(drawed)]} {
+        unset rectangle(drawed)
+	set rectangle(x2) [$c canvasx $x]
+	set rectangle(y2) [$c canvasy $y]
+
+        set id [create_id]
+	set tag [rectangle_tag $id]
+
+        catch {$c delete $rectangle(temp)}
+        set rectangle(temp) [svg::parseSVGItem $c {} {} $tag]
+
+	if {[chat::is_groupchat $chatid]} {
+	    $c delete $rectangle(temp)
+	} else {
+	    set jid [jlib::connection_jid [chat::get_connid $chatid]]
+	    $c addtag [list tag $tag] withtag $rectangle(temp)
+	    $c addtag [time_tag created $jid] withtag $rectangle(temp)
+	}
+
+	send_svg $chatid $tag
+
+        unset rectangle(x1)
+        unset rectangle(x2)
+        unset rectangle(y1)
+        unset rectangle(y2)
+        set rectangle(temp) {}
+    }
+}
+
+proc wb::rectangle_b1m {c x y} {
+    variable rectangle
+
+    if {[info exists rectangle(drawed)]} {
+	set rectangle(x2) [$c canvasx $x]
+	set rectangle(y2) [$c canvasy $y]
+
+	set tag [rectangle_tag ""]
+
+	catch {$c delete $rectangle(temp)}
+        set rectangle(temp) [svg::parseSVGItem $c {} {} $tag]
+    }
+}
+
+proc wb::rectangle_tag {id} {
+    variable rectangle
+
+    set vars $rectangle(options)
+    if {$rectangle(x2) > $rectangle(x1)} {
+	lappend vars x $rectangle(x1) \
+		     width [expr {$rectangle(x2) - $rectangle(x1)}]
+    } else {
+	lappend vars x $rectangle(x2) \
+		     width [expr {$rectangle(x1) - $rectangle(x2)}]
+    }
+    if {$rectangle(y2) > $rectangle(y1)} {
+	lappend vars y $rectangle(y1) \
+		     height [expr {$rectangle(y2) - $rectangle(y1)}]
+    } else {
+	lappend vars y $rectangle(y2) \
+		     height [expr {$rectangle(y1) - $rectangle(y2)}]
+    }
+    if {$id != ""} {
+	lappend vars id $id
+    }
+    return [jlib::wrapper:createtag rect -vars $vars]
+}
+
 ###############################################################################
 
 ###############################################################################
 # Circle
 
-proc wb::circle_bind {c jid} {
+proc wb::circle_bind {c chatid} {
     bind $c <ButtonPress-1> \
-	[list [namespace current]::circle_b1 [double% $c] [double% $jid] %x %y]
+	[list [namespace current]::circle_b1 [double% $c] [double% $chatid] %x %y]
     bind $c <B1-Motion> \
  	[list [namespace current]::circle_b1m [double% $c] %x %y]
     bind $c <Motion> {}
     bind $c <ButtonRelease-1> \
-	[list [namespace current]::circle_b1r [double% $c] [double% $jid] %x %y]
+	[list [namespace current]::circle_b1r [double% $c] [double% $chatid] %x %y]
     bind $c <Button-3> {}
+    $c configure -cursor crosshair
 }
 
-proc wb::circle_b1 {c jid x y} {
+proc wb::circle_b1 {c chatid x y} {
     variable circle
     global dofill
 
-    set x [$c canvasx $x]
-    set y [$c canvasy $y]
+    set cx [$c canvasx $x]
+    set cy [$c canvasy $y]
 
     set circle(drawed) 1
-    set circle(center) "$x $y"
-    set circle(options) [list -outline [get_color $jid] \
-			       -width [get_width $jid] ]
+    set circle(cx) $cx
+    set circle(cy) $cy
+    set circle(options) [list cx $cx cy $cy stroke [get_color $chatid]]
     if {$dofill == 1} {
-	lappend circle(options) -fill [get_fill_color $jid]
+	lappend circle(options) fill [get_fill_color $chatid]
+    }
+    if {[set width [get_width $chatid]] != 1} {
+	lappend circle(options) stroke-width $width
     }
 }
 
 proc wb::circle_b1r {c chatid x y} {
     variable circle
-    global dofill
 
     if {[info exists circle(drawed)]} {
         unset circle(drawed)
-	set cx [lindex $circle(center) 0]
-	set cy [lindex $circle(center) 1]
+	set cx $circle(cx)
+	set cy $circle(cy)
 	set x [$c canvasx $x]
 	set y [$c canvasy $y]
-	set r [expr hypot($cx-$x,$cy-$y)]
-	set x1 [expr $cx-$r]
-	set x2 [expr $cx+$r]
-	set y1 [expr $cy-$r]
-	set y2 [expr $cy+$r]
+	set r [expr {hypot($cx - $x, $cy - $y)}]
 
         set id [create_id]
-        lappend vars cx $cx
-	lappend vars cy $cy
-	lappend vars r $r
-        if {[set color [get_color $chatid]] != "#000000"} {
-	   lappend vars stroke $color
-	}
-	if {$dofill == 1} {
-	   lappend vars fill [get_fill_color $chatid]
-	}
-	if {[set width [get_width $chatid]] != 1} {
-	   lappend vars stroke-width $width
-	}
-
-        lappend vars id $id
+	set tag [circle_tag $id $r]
 
         catch {$c delete $circle(temp)}
-        set circle(temp) [eval $c create oval $x1 $y1 $x2 $y2 $circle(options) \
-				  -tag id$id]
-	    
+        set circle(temp) [svg::parseSVGItem $c {} {} $tag]
+
 	if {[chat::is_groupchat $chatid]} {
 	    $c delete $circle(temp)
 	} else {
 	    set jid [jlib::connection_jid [chat::get_connid $chatid]]
+	    $c addtag [list tag $tag] withtag $circle(temp)
 	    $c addtag [time_tag created $jid] withtag $circle(temp)
 	}
 
-	send_svg $chatid [jlib::wrapper:createtag circle -vars $vars]
+	send_svg $chatid $tag
 
-        set circle(center) {}
+        unset circle(cx)
+        unset circle(cy)
         set circle(temp) {}
     }
 }
@@ -699,58 +944,58 @@ proc wb::circle_b1m {c x y} {
     variable circle
 
     if {[info exists circle(drawed)]} {
-	set cx [lindex $circle(center) 0]
-	set cy [lindex $circle(center) 1]
+	set cx $circle(cx)
+	set cy $circle(cy)
 	set x [$c canvasx $x]
 	set y [$c canvasy $y]
-	set r [expr hypot($cx-$x,$cy-$y)]
-	set x1 [expr $cx-$r]
-	set x2 [expr $cx+$r]
-	set y1 [expr $cy-$r]
-	set y2 [expr $cy+$r]
-		
+	set r [expr hypot($cx - $x, $cy - $y)]
+
+	set tag [circle_tag "" $r]
 
 	catch {$c delete $circle(temp)}
-	set circle(temp) [eval $c create oval $x1 $y1 $x2 $y2 $circle(options)]
+        set circle(temp) [svg::parseSVGItem $c {} {} $tag]
     }
 }
 
+proc wb::circle_tag {id r} {
+    variable circle
 
+    set vars $circle(options)
+    lappend vars r $r
+    if {$id != ""} {
+	lappend vars id $id
+    }
+    return [jlib::wrapper:createtag circle -vars $vars]
+}
 
 ###############################################################################
 
 ###############################################################################
 # Freehand
 
-proc wb::freehand_bind {c jid} {
+proc wb::freehand_bind {c chatid} {
     bind $c <ButtonPress-1> \
 	[list [namespace current]::freehand_b1p \
-	     [double% $c] [double% $jid] %x %y]
+	     [double% $c] [double% $chatid] %x %y]
     bind $c <B1-Motion> [list [namespace current]::freehand_b1m \
 			     [double% $c] %x %y]
     bind $c <ButtonRelease-1> \
-	[list [namespace current]::freehand_b1r [double% $c] [double% $jid]]
+	[list [namespace current]::freehand_b1r [double% $c] [double% $chatid]]
     bind $c <Button-3> {}
+    $c configure -cursor crosshair
 }
 
-proc wb::freehand_b1p {c jid x y} {
+proc wb::freehand_b1p {c chatid x y} {
     variable line
 
     set x [$c canvasx $x]
     set y [$c canvasy $y]
-    #puts "$x $y"
 
-    if {0 && [info exists line(drawed)]} {
-	lappend line(coords) $x $y
-
-	catch {$c delete $line(temp)}
-	set line(temp) [eval $c create line $line(coords)]
-    } else {
-	set line(drawed) 1
-	set line(coords) "$x $y"
-	set line(options) [list -fill [get_color $jid] \
-			       -width [get_width $jid] \
-			       -joinstyle round]
+    set line(drawed) 1
+    set line(coords) [list $x $y]
+    set line(options) [list stroke-linejoin round stroke [get_color $chatid]]
+    if {[set width [get_width $chatid]] != 1} {
+	lappend line(options) stroke-width $width
     }
 }
 
@@ -764,7 +1009,8 @@ proc wb::freehand_b1m {c x y} {
 	lappend line(coords) $x $y
 
 	catch {$c delete $line(temp)}
-	set line(temp) [eval $c create line $line(coords) $x $y $line(options)]
+	set tag [freehand_tag ""]
+	set line(temp) [svg::parseSVGItem $c {} {} $tag]
     }
 }
 
@@ -776,65 +1022,51 @@ proc wb::freehand_b1r {c chatid} {
 	unset line(drawed)
 
 	set id [create_id]
+
+	set tag [freehand_tag $id]
+
 	catch {$c delete $line(temp)}
-	set line(temp) [eval $c create line $line(coords) $line(options) \
-			    -tag id$id]
+	set line(temp) [svg::parseSVGItem $c {} {} $tag]
 
 	if {[chat::is_groupchat $chatid]} {
 	    $c delete $line(temp)
 	} else {
 	    set jid [jlib::connection_jid [chat::get_connid $chatid]]
+	    $c addtag [list tag $tag] withtag $line(temp)
 	    $c addtag [time_tag created $jid] withtag $line(temp)
 	}
 
-	lappend vars points $line(coords)
-	if {[set color [get_color $chatid]] != "#000000"} {
-	    lappend vars stroke $color
-	}
-	if {[set width [get_width $chatid]] != 1} {
-	    lappend vars stroke-width $width
-	}
-
-	lappend vars id $id
-
-	send_svg $chatid [jlib::wrapper:createtag polyline \
-			   -vars $vars]
+	send_svg $chatid $tag
 
 	set line(coords) {}
 	set line(temp) {}
     }
 }
 
+proc wb::freehand_tag {id} {
+    variable line
+
+    set vars $line(options)
+    lappend vars points $line(coords)
+    if {$id != ""} {
+	lappend vars id $id
+    }
+    return [jlib::wrapper:createtag polyline -vars $vars]
+}
+
 ###############################################################################
 
 ###############################################################################
 # Remove
 
-proc wb::remove_bind {c jid} {
-    bind $c <ButtonPress-1> \
-	[list [namespace current]::remove_b1p \
-	     [double% $c] [double% $jid]]
-    bind $c <B1-Motion> {}
-    bind $c <ButtonRelease-1> {}
-    bind $c <Button-3> {}
-}
-
-proc wb::remove_b1p {c chatid} {
-    set tags [$c gettags current]
-    set id ""
-    foreach t $tags {
-	if {[crange $t 0 1] == "id"} {
-	    set id [crange $t 2 end]
-	    break
-	}
-    }
+proc wb::remove_b1p {c chatid id} {
     set connid [chat::get_connid $chatid]
     set jid [chat::get_jid $chatid]
     if {[chat::is_groupchat $chatid]} {
 	set type groupchat
     } else {
 	set type chat
-	$c delete id$id
+	$c delete [list id $id]
     }
     if {$id != ""} {
 	jlib::send_msg $jid \
@@ -844,7 +1076,6 @@ proc wb::remove_b1p {c chatid} {
 			      -vars {xmlns tkabber:whiteboard} \
 			      -subtags [list [jlib::wrapper:createtag remove \
 						  -vars [list id $id]]]]]
-	
     }
 }
 
@@ -854,25 +1085,27 @@ proc wb::remove_b1p {c chatid} {
 ###############################################################################
 # Move
 
-proc wb::move_bind {c jid} {
+proc wb::move_bind {c chatid} {
     bind $c <ButtonPress-1> \
-	[list [namespace current]::move_b1p \
-	     [double% $c] [double% $jid] %x %y]
-    bind $c <B1-Motion> [list [namespace current]::move_b1m \
-			     [double% $c] %x %y]
+	 [list [namespace current]::move_b1p [double% $c] [double% $chatid] %x %y]
+    bind $c <B1-Motion> \
+	 [list [namespace current]::move_b1m [double% $c] %x %y]
     bind $c <ButtonRelease-1> \
-	[list [namespace current]::move_b1r [double% $c] [double% $jid]]
-    bind $c <Button-3> {}
+	 [list [namespace current]::move_b1r [double% $c] [double% $chatid]]
+    bind $c <Button-3> \
+	 [list [namespace current]::popup_menu [double% $c] [double% $chatid] %X %Y]
+    bind $c <Motion> {}
+    $c configure -cursor ""
 }
 
-proc wb::move_b1p {c jid x y} {
+proc wb::move_b1p {c chatid x y} {
     variable move
 
     set tags [$c gettags current]
     set id ""
     foreach t $tags {
-	if {[crange $t 0 1] == "id"} {
-	    set id [crange $t 2 end]
+	if {[lindex $t 0] == "id"} {
+	    set id [lindex $t 1]
 	    break
 	}
     }
@@ -886,6 +1119,7 @@ proc wb::move_b1p {c jid x y} {
 	set move(lasty) $y
 
 	set move(id) $id
+	$c configure -cursor hand2
     } else {
 	catch {unset move(id)}
     }
@@ -899,7 +1133,7 @@ proc wb::move_b1m {c x y} {
 	set x [$c canvasx $x]
 	set y [$c canvasy $y]
 
-	$c move id$id [expr {$x - $move(lastx)}] [expr {$y - $move(lasty)}]
+	$c move [list id $id] [expr {$x - $move(lastx)}] [expr {$y - $move(lasty)}]
 
 	set move(lastx) $x
 	set move(lasty) $y
@@ -915,19 +1149,24 @@ proc wb::move_b1r {c chatid} {
 	set x $move(lastx)
 	set y $move(lasty)
 
+	set dx [expr {$x - $move(startx)}]
+	set dy [expr {$y - $move(starty)}]
+
+	if {$dx == 0 && $dy == 0} {
+	    $c configure -cursor ""
+	    return
+	}
+
 	if {[chat::is_groupchat $chatid]} {
 	    set type groupchat
-	    $c move id$id \
-		[expr {$move(startx) - $x}] [expr {$move(starty) - $y}]
+	    $c move [list id $id] [expr {-$dx}] [expr {-$dy}]
 	} else {
 	    set type chat
 	    set jid [jlib::connection_jid [chat::get_connid $chatid]]
-	    $c addtag [time_tag moved $jid] withtag id$id
+	    $c addtag [time_tag moved $jid] withtag [list id $id]
 	}
 
-	set vars [list id $id \
-		      dx [expr {$x - $move(startx)}] \
-		      dy [expr {$y - $move(starty)}]]
+	set vars [list id $id dx $dx dy $dy]
 
 	set connid [chat::get_connid $chatid]
 	set jid [chat::get_jid $chatid]
@@ -938,6 +1177,7 @@ proc wb::move_b1r {c chatid} {
 			      -vars {xmlns tkabber:whiteboard} \
 			      -subtags [list [jlib::wrapper:createtag move \
 						  -vars $vars]]]]
+	$c configure -cursor ""
     }
 }
 
@@ -950,12 +1190,13 @@ proc wb::text_bind {c chatid} {
     bind $c <ButtonRelease-1> \
 	    [list [namespace current]::text_b1 [double% $c] [double% $chatid] %x %y]
     bind $c <Button-3> {}
+    $c configure -cursor crosshair
 }
 
 proc wb::text_b1 {c chatid x y} {
     variable text_info
-    set text_info(x) $x
-    set text_info(y) $y
+    set text_info(x) [$c canvasx $x]
+    set text_info(y) [$c canvasy $y]
     set w [win_id whiteboard $chatid]
     set wt $w.text_dialog
     if {[winfo exists $wt]} {
@@ -971,7 +1212,7 @@ proc wb::text_b1 {c chatid x y} {
 	$wt add -text [::msgcat::mc "OK"] \
 	    -command [list [namespace current]::text_ok $wt $c $chatid]
 	$wt add -text [::msgcat::mc "Cancel"] \
-	    -command [list wm withdraw $wt]
+	    -command [list [namespace current]::text_cancel $wt $c $chatid]
 
 	set en [entry $wt.text -width 80 \
 		    -textvariable [namespace current]::text_entered($chatid)]
@@ -986,20 +1227,7 @@ proc wb::text_ok {wt c chatid} {
     set id [create_id]
 
     set text [set [namespace current]::text_entered($chatid)]
-    if {[chat::is_groupchat $chatid]} {
-	set type groupchat
-    } else {
-	set type chat
-	$c create text $text_info(x) $text_info(y) -tag id$id \
-	    -text $text \
-	    -fill [get_text_color $chatid]
-	set font [get_text_font $chatid]
-	if {[info exists app_font($font)]} {
-	    $c itemconfigure id$id -font $font
-	}
-	set jid [jlib::connection_jid [chat::get_connid $chatid]]
-	$c addtag [time_tag created $jid] withtag id$id
-    }
+
     set vars [list id $id x $text_info(x) y $text_info(y) \
 		   fill [get_text_color $chatid]]
     set font [get_text_font $chatid]
@@ -1026,10 +1254,22 @@ proc wb::text_ok {wt c chatid} {
 	unset font_opt
     }
 
-    send_svg $chatid [jlib::wrapper:createtag text -vars $vars -chdata $text]
+    set tag [jlib::wrapper:createtag text -vars $vars -chdata $text]
+
+    if {![chat::is_groupchat $chatid]} {
+	set textid [svg::parseSVGItem $c {} {} $tag]
+	set jid [jlib::connection_jid [chat::get_connid $chatid]]
+	$c addtag [list tag $tag] withtag $textid
+	$c addtag [time_tag created $jid] withtag $textid
+    }
+
+    send_svg $chatid $tag
     wm withdraw $wt
 }
 
+proc wb::text_cancel {wt c chatid} {
+    wm withdraw $wt
+}
 
 ###############################################################################
 
